{{#if reference}}----- Start of reference content -----
{{reference}}
----- End of reference content -----
{{/if}}
## Content to Check
----- Start of content to check -----
{{to_check}}
----- End of content to check -----

## Rule to Follow
{{rule}}

## Judge thinks `Content to check` does violate the `Rule to Follow`
{{judge}}

## Improvement Instructions
1. Analyze the content against the rule
2. Identify specific violations or issues
3. For each issue, propose a clear, actionable solution